<?php
/************************************************************************
 * MXChange v0.2.1                                    Start: 03/25/2004 *
 * ===============                              Last change: 09/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : extensions.php                                   *
 * -------------------------------------------------------------------- *
 * Short description : Extension management                             *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Erweiterungen-Management                         *
 * -------------------------------------------------------------------- *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2008 by Roland Haeder                           *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	$INC = substr(dirname(__FILE__), 0, strpos(dirname(__FILE__), "/inc") + 4) . "/security.php";
	require($INC);
}

// Load the extension and maybe found language and function files.
function LOAD_EXTENSION ($ext_name, $EXT_LOAD_MODE = "") {
	global $EXT_LOADED, $_CONFIG, $CSS, $cacheMode;

	// Is the extension already loaded?
	//* DEBUG: */ echo "Loading extension {$ext_name}.<br />\n";
	if (isset($EXT_LOADED[$ext_name])) {
		// Debug message
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Extension %s already loaded.", $ext_name));
		return false;
	}

	// Construct FQFN for extension file
	$extInclude = sprintf("%sinc/extensions/ext-%s.php", PATH, $ext_name);

	// Is the extension file NOT there?
	if (!FILE_READABLE($extInclude)) {
		// Debug message
		DEBUG_LOG(__FUNCTION__, __LINE__, sprintf("Extension %s not found.", $ext_name));

		// Abort here
		return false;
	} // END - if

	// Construct FQFN for language file
	$langInclude = sprintf("%sinc/language/%s_%s.php", PATH, $ext_name, GET_LANGUAGE());

	// Is this include there?
	if (FILE_READABLE($langInclude)) {
		// Then load it
		//* DEBUG: */ echo "--- Language loaded.<br />\n";
		require($langInclude);
	} // END - if

	// Construct FQFN for functions file
	$funcsInclude = sprintf("%sinc/libs/%s_functions.php", PATH, $ext_name);

	// Is this include there?
	if (FILE_READABLE($funcsInclude)) {
		// Then load it
		//* DEBUG: */ echo "--- Functions loaded.<br />\n";
		require($funcsInclude);
	} // END - if

	// Extensions are not deprecated by default
	$EXT_DEPRECATED = "N";

	// Include the extension file
	//* DEBUG: */ echo "Extension loaded.<br />\n";
	require($extInclude);

	// Is this extension deprecated?
	if ($EXT_DEPRECATED == "Y") {
		// Deactivate the extension
		DEACTIVATE_EXTENSION($ext_name);

		// Abort here
		return false;
	} // END - if

	// Mark it as loaded
	$EXT_LOADED[$ext_name] = true;

	// All fine!
	return true;
}

//
function EXTENSION_REGISTER ($ext_name, $id, $dry_run=false) {
	global $NOTES, $_CONFIG, $INC_POOL, $cacheInstance;

	// This shall never do a non-admin user!
	if (!IS_ADMIN()) return false;

	// Is this extension already installed?
	if (EXT_IS_ACTIVE($ext_name)) return false;

	// We want to register an extension and registration status is by default "failed" (= false)
	$EXT_LOAD_MODE = "register"; $ret = false; $SQLs = array();
	$INC_POOL = array();

	// By default the language prefix is the extension's name
	// @TODO: Do we really need this one anymore? Can't we just take $ext_name and done?
	// By default we have no failtures
	$EXT_REPORTS_FAILURE = false;

	// Generate file name
	$file = sprintf("%sinc/extensions/ext-%s.php", PATH, $ext_name);

	// Does this extension exists?
	if (FILE_READABLE($file)) {
		// Extension was found so we can load it in registration mode
		$EXT_ALWAYS_ACTIVE = "N";
		require($file);

		// And run possible updates
		$EXT_LOAD_MODE = "update"; $EXT_UPDATE_DEPENDS = "";
		foreach ($EXT_VER_HISTORY as $EXT_VER) {
			// Load extension in update mode
			require($file);

			// Do we have an update?
			if (((GET_EXT_VERSION("sql_patches") != "") && ($_CONFIG['verbose_sql'] == "Y")) || (!EXT_IS_ACTIVE("sql_patches"))) {
				if (!empty($UPDATE_NOTES)) {
					// Update notes found
					$NOTES .= "<STRONG>v".$EXT_VER.":</STRONG><br />".$UPDATE_NOTES."<br /><br />\n";
					$UPDATE_NOTES = "";
				} elseif (($EXT_VER == "0.0") || ($EXT_VER == "0.0.0")) {
					// Initial release
					$NOTES .= "<STRONG>v".$EXT_VER.":</STRONG><br />".INITIAL_RELEASE."<br /><br />\n";
				} else {
					// No update notes found!
					$NOTES .= "<STRONG>v".$EXT_VER.":</STRONG><br /><I>".NO_UPDATE_NOTES."</I><br /><br />\n";
				}
			} // END - if
		} // END - foreach

		// Does this extension depends on an outstanding update of another update?
		if (!empty($EXT_UPDATE_DEPENDS)) {
			// Backup SQL commands and clear current
			$SQLs2 = $SQLs;
			$SQLs  = array();
			$test  = false;

			// Load required extension also in update mode
			$file = sprintf("%sinc/extensions/ext-%s.php", PATH, $EXT_UPDATE_DEPENDS);

			// Check for required file
			if (FILE_READABLE($file)) {
				// Bacup version number
				$VER_BACKUP = $EXT_VERSION;

				// Save the Parrent $EXT_ALWAYS_ACTIVE for later!
				$EXT_ALWAYS_ACTIVE_PARRENT = $EXT_ALWAYS_ACTIVE;

				// Set EXT_ALWAYS_ACTIVE for update
				$EXT_ALWAYS_ACTIVE = "N";

				// File exists so let's load it
				require($file);

				// If versions mismatch update extension first
				$ext_ver = GET_EXT_VERSION($EXT_UPDATE_DEPENDS);

				// Extension version set? If empty the extension is not registered
				if (empty($ext_ver)) {
					// Extension not registered so far so first load task's ID...
					$result = SQL_QUERY_ESC("SELECT id FROM "._MYSQL_PREFIX."_task_system WHERE task_type='EXTENSION' AND subject LIKE '[%s:]%%' LIMIT 1",
					 array($EXT_UPDATE_DEPENDS), __FILE__, __LINE__);

					// Entry found?
					if (SQL_NUMROWS($result) == 1) {
						// Task found so load task's ID and register extension...
						list($task) = SQL_FETCHROW($result);

						// Try to register the extension
						$test = EXTENSION_REGISTER($EXT_UPDATE_DEPENDS, $task, $dry_run);
					} // END - if

					// Free result
					SQL_FREERESULT($result);
				} elseif ($ext_ver != $EXT_VERSION) {
					// Ok, update this extension now
					EXTENSION_UPDATE(basename($file), $EXT_UPDATE_DEPENDS, $ext_ver, $dry_run);

					// All okay!
					$test = true;
				} else {
					// Nothing to register / update before...
					$test = true;
				}

				// Restore version number
				$EXT_VERSION = $VER_BACKUP;

				// Restore $EXT_ALWAYS_ACTIVE with the value from parrent
				$EXT_ALWAYS_ACTIVE = $EXT_ALWAYS_ACTIVE_PARRENT;
			} else {
				// Required file for update does not exists!
				$test = true;
				// But this is fine for the first time...
			}

			// Finally restore previous SQLs
			$SQLs = $SQLs2; unset($SQLs2);
		} else {
			// Does not depend on an other extension
			$test = true;
		}

		// Switch back to register mode
		$EXT_LOAD_MODE = "register";

		// Remains true if extension registration reports no failtures
		$test = ($test && !$EXT_REPORTS_FAILURE);

		// Does everthing before wents ok?
		if ($test) {
			// "Dry-run-mode" activated?
			if (!$dry_run) {
				// Run all SQLs
				foreach ($SQLs as $sql) {
					// Trim spaces away which we don't need
					$sql = trim($sql);

					// Is there still an SQL query?
					if (!empty($sql)) {
						// Do we have an "ALTER TABLE" command?
						if (substr(strtolower($sql), 0, 11) == "alter table") {
							// Analyse the alteration command
							SQL_ALTER_TABLE($sql, __FILE__, __LINE__);
						} else {
							// Run regular SQL command
							$result = SQL_QUERY($sql, __FILE__, __LINE__, false);
						}
					} // END - if
				} // END - foreach

				// Remove cache file(s) if extension is active
				if ((EXT_IS_ACTIVE("cache")) || (GET_EXT_VERSION("cache") != "")) {
					//* DEBUG: */ echo __LINE__.": DESTROY!<br />\n";
					// Remove cache files
					if ($cacheInstance->loadCacheFile("extensions", true)) $cacheInstance->destroyCacheFile();
					if ($cacheInstance->loadCacheFile("mod_reg"))          $cacheInstance->destroyCacheFile();
					if ($cacheInstance->loadCacheFile("config"))           $cacheInstance->destroyCacheFile();
				} // END - if

				// Check for added include files
				if (count($INC_POOL > 0)) {
					// Loads every include file
					foreach ($INC_POOL as $inc) {
						require_once($inc);
					} // END - foreach

					// Remove array
					unset($INC_POOL);
				} // END - if

				// Register extension
				$result = SQL_QUERY_ESC("INSERT INTO "._MYSQL_PREFIX."_extensions (ext_name, ext_active, ext_version) VALUES ('%s','%s','%s')",
				 array($ext_name, $EXT_ALWAYS_ACTIVE, $EXT_VERSION), __FILE__, __LINE__);

				// Update task management
				ADMIN_SOLVE_TASK($id);

				// @TODO This causes the whole (!) menu cache being purged
				CACHE_PURGE_ADMIN_MENU();

				// In normal mode return a true on success
				$ret = true;

				// Remove SQLs
				unset($SQLs);
			} else {
				// Rewrite SQL command to keep { and } inside
				foreach ($SQLs as $key => $sql) {
					$sql = str_replace('{', "&#123;", str_replace('}', "&#125;", $sql));
					$SQLs[$key] = $sql;
				} // END - foreach

				// In  "dry-run" mode return array with all SQL commands
				$ret = $SQLs;

				// Remove all SQL commands
				unset($SQLs);
			}
		} else {
			// No, an error occurs while registering extension :-(
			$ret = false;
		}
	} elseif (($id > 0) && (!empty($ext_name))) {
		// Remove task from system when id and extension's name is valid
		$result = SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM "._MYSQL_PREFIX."_task_system WHERE id=%s AND status='NEW' LIMIT 1",
		 array(bigintval($id)), __FILE__, __LINE__);
	}

	// Is this the sql_patches?
	//* DEBUG: */ echo __LINE__.":{$ext_name}/{$EXT_LOAD_MODE}<br />\n";
	if (($ext_name == "sql_patches") && (($EXT_LOAD_MODE == "register") || ($EXT_LOAD_MODE == "remove")) && (!$dry_run) && ($test)) {
		// Then redirect to logout
		//* DEBUG: */ echo __LINE__.": LOAD!<br />\n";
		LOAD_URL("modules.php?module=admin&logout=1&".$EXT_LOAD_MODE."=sql_patches");
	} // END - if

	// Return status code
	return $ret;
}
//
function EXTENSION_RUN_SQLS($id, $EXT_LOAD_MODE) {
	global $cacheInstance, $_CONFIG;
	$SQLs = array();

	// Extensions are never active by default
	$EXT_ALWAYS_ACTIVE = "N";

	// By default no SQL has been executed
	$sqlRan = false;

	// This shall never do a non-admin user!
	if (!IS_ADMIN()) return false;

	// Get extension's name
	$ext_name = GET_EXT_NAME($id);
	if (empty($ext_name)) return false;

	// Load extension in detected mode
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):ext_name[{$id}]={$ext_name}<br />\n";
	$file = sprintf("%sinc/extensions/ext-%s.php", PATH, $ext_name);
	if (FILE_READABLE($file)) {
		// Load the include
		require($file);
	} // END - if

	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):SQLs::count=".count($SQLs)."<br />\n";
	if ((is_array($SQLs) && (sizeof($SQLs) > 0))) {
		// Run SQL commands...
		foreach ($SQLs as $sql) {
			// Trim spaces away which we don't need
			$sql = trim($sql);

			// Is there still an SQL query?
			if (!empty($sql)) {
				// Do we have an "ALTER TABLE" command?
				//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):SQL={$SQL}<br />\n";
				if (substr(strtolower($sql), 0, 11) == "alter table") {
					// Analyse the alteration command
					SQL_ALTER_TABLE($sql, __FILE__, __LINE__);
				} else {
					// Run regular SQL command
					$result = SQL_QUERY($sql, __FILE__, __LINE__, false);
				}

				// An SQL has been executed
				$sqlRan = true;
			} // END - if
		} // END - foreach

		// Removal mode?
		if ($EXT_LOAD_MODE == "remove") {
			// Delete this extension (remember to remove it from your server *before* you click on welcome!
			$result = SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM "._MYSQL_PREFIX."_extensions WHERE id=%s LIMIT 1",
			 array($id), __FILE__, __LINE__);
		} // END - if

		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):mode={$EXT_LOAD_MODE}<br />\n";

		// Is this the sql_patches?
		//* DEBUG: */ echo __LINE__.": {$id}/{$ext_name}/{$EXT_LOAD_MODE}<br />\n";
		if (($ext_name == "sql_patches") && (($EXT_LOAD_MODE == "register") || ($EXT_LOAD_MODE == "remove"))) {
			// Then redirect to logout
			//* DEBUG: */ echo __LINE__.": LOAD!<br />\n";
			LOAD_URL("modules.php?module=admin&logout=1&".$EXT_LOAD_MODE."=sql_patches");
		} // END - if
	} // END - if

	// Remove cache file(s) if extension is active
	if (((EXT_IS_ACTIVE("cache")) || (GET_EXT_VERSION("cache") != "")) && (((SQL_AFFECTEDROWS() == 1)) || ($sqlRan === true) || ($EXT_LOAD_MODE == "activate") || ($EXT_LOAD_MODE == "deactivate"))) {
		//* DEBUG: */ echo __LINE__.": DESTROY!<br />\n";
		// Remove cache files
		if ($cacheInstance->loadCacheFile("extensions", true)) $cacheInstance->destroyCacheFile();
		if ($cacheInstance->loadCacheFile("mod_reg"))    $cacheInstance->destroyCacheFile();
		if ($cacheInstance->loadCacheFile("config"))     $cacheInstance->destroyCacheFile();

		// @TODO This causes the whole (!) menu cache being purged
		CACHE_PURGE_ADMIN_MENU();
	} // END - if
}
// Check if given extension is active
function EXT_IS_ACTIVE ($ext_name) {
	global $cacheArray, $_CONFIG;

	// Extensions are all inactive during installation
	if ((!isBooleanConstantAndTrue('mxchange_installed')) || (isBooleanConstantAndTrue('mxchange_installing')) || (empty($ext_name))) return false;

	// Not active is the default
	$active = "N";

	// Check cache
	if (!empty($cacheArray['extensions']['ext_active'][$ext_name])) {
		// Load from cache
		//* DEBUG: */ echo "CACHE! ext_name={$ext_name}<br />\n";
		$active = $cacheArray['extensions']['ext_active'][$ext_name];

		// Count cache hits
		if (isset($_CONFIG['cache_hits'])) $_CONFIG['cache_hits']++;
	} elseif (($ext_name == "cache") || (GET_EXT_VERSION("cache") == "")) {
		//* DEBUG: */ echo "DB! ext_name={$ext_name}<br />\n";
		// Load from database
		$result = SQL_QUERY_ESC("SELECT ext_active FROM "._MYSQL_PREFIX."_extensions WHERE ext_name='%s' LIMIT 1",
		 array($ext_name), __FILE__, __LINE__);
		if (SQL_NUMROWS($result) == 0) {
			// Extension was not found!
			return false;
		}

		// Load entry
		list($active) = SQL_FETCHROW($result);

		// Free result
		SQL_FREERESULT($result);


		// Write cache array
		//* DEBUG: */ echo $ext_name."[DB]: {$active}<br />\n";
		$cacheArray['extensions']['ext_active'][$ext_name] = $active;
	} else {
		// Extension not active!
		//* DEBUG: */ echo $ext_name.": Not active!<br />\n";
		$cacheArray['extensions']['ext_active'][$ext_name] = "N";
	}

	// Debug message
	//* DEBUG: */ DEBUG_LOG(__FUNCTION__, __LINE__, " ext_name={$ext_name},active={$active}");

	// Is this extension activated? (For admins we always have active extensions...)
	return ($active == "Y");
}
// Get version from extensions
function GET_EXT_VERSION ($ext_name) {
	global $cacheArray, $_CONFIG, $cacheInstance;
	$ret = false;

	// Extensions are all inactive during installation
	if ((!isBooleanConstantAndTrue('mxchange_installed')) || (isBooleanConstantAndTrue('mxchange_installing'))) return "";
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ext_name={$ext_name}<br />\n";

	// Is the cache written?
	if (!empty($cacheArray['extensions']['ext_version'][$ext_name])) {
		// Load data from cache
		//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): CACHE!<br />\n";
		$ret = $cacheArray['extensions']['ext_version'][$ext_name];

		// Count cache hits
		if (isset($_CONFIG['cache_hits'])) $_CONFIG['cache_hits']++; else $_CONFIG['cache_hits'] = 1;
	} elseif (!is_object($cacheInstance)) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT ext_version FROM "._MYSQL_PREFIX."_extensions WHERE ext_name='%s' LIMIT 1",
		 array($ext_name), __FILE__, __LINE__);
		list($ret) = SQL_FETCHROW($result);
		SQL_FREERESULT($result);

		// Set cache
		$cacheArray['extensions']['ext_version'][$ext_name] = $ret;
	}

	// Return result
	//* DEBUG: */ print __FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ret={$ret}<br />\n";
	return $ret;
}
//
function EXTENSION_UPDATE($file, $ext, $EXT_VER, $dry_run=false) {
	// This shall never do a non-admin user!
	global $cacheInstance, $_CONFIG, $NOTES;

	// Init arrays
	$SQLs = array(); $INC_POOL = array();

	// Only admins are allowed to update extensions
	if ((!IS_ADMIN()) || (empty($ext))) return false;

	// Load extension in update mode
	$EXT_LOAD_MODE = "update"; $EXT_UPDATE_DEPENDS = ""; $NOTES = "";

	// Load extension file
	include(sprintf("%sinc/extensions/%s", PATH, $file));

	if (!empty($EXT_UPDATE_DEPENDS)) {
		// Update another extension first!
		$test = EXTENSION_UPDATE(("ext-".$EXT_UPDATE_DEPENDS.".php"), $EXT_UPDATE_DEPENDS, GET_EXT_VERSION($EXT_UPDATE_DEPENDS), $dry_run);
	}

	// Check if version is updated
	if ((($EXT_VERSION != $EXT_VER) || ($dry_run)) && (is_array($EXT_VER_HISTORY))) {
		// Search for starting point
		$start = array_search($EXT_VER, $EXT_VER_HISTORY);
		$NOTES = "";

		// And load SQL queries in order of version history
		for ($idx = ($start + 1); $idx < sizeof($EXT_VER_HISTORY); $idx++) {
			// Remove old SQLs array to prevent possible bugs
			if (!$dry_run) { unset($SQLs); $SQLs = array(); }

			// Set version
			$EXT_VER = $EXT_VER_HISTORY[$idx];

			// Include again...
			include(PATH."inc/extensions/".$file);

			// Add notes
			if ($_CONFIG['verbose_sql'] == "Y") {
				$EXT_VER = $EXT_VER_HISTORY[$idx];
				if (!empty($UPDATE_NOTES)) {
					// Update notes found
					$NOTES .= "<STRONG>v".$EXT_VER.":</STRONG><br />".$UPDATE_NOTES."<br /><br />\n";
					$UPDATE_NOTES = "";
				} elseif ($EXT_VER == "0.0") {
					// Initial release
					$NOTES .= "<STRONG>v".$EXT_VER.":</STRONG><br />".INITIAL_RELEASE."<br /><br />\n";
				} else {
					$NOTES .= "<STRONG>v".$EXT_VER.":</STRONG><br /><I>".NO_UPDATE_NOTES."</I><br /><br />\n";
				}
			} // END - if

			// In real-mode execute any existing includes
			if ((!$dry_run) && (count($INC_POOL) > 0)) {
				// Include all files
				foreach ($INC_POOL as $fqfn) {
					require_once($fqfn);
				} // END - foreach
			} // END - if

			// Run SQLs
			if ((is_array($SQLs)) && (!$dry_run)) {
				// Run SQL commands
				foreach ($SQLs as $sql)
				{
					$sql = trim($sql);
					if (!empty($sql))
					{
						// Do we have an "ALTER TABLE" command?
						if (substr(strtolower($sql), 0, 11) == "alter table") {
							// Analyse the alteration command
							SQL_ALTER_TABLE($sql, __FILE__, __LINE__);
						} else {
							// Run regular SQL command
							$result = SQL_QUERY($sql, __FILE__, __LINE__, false);
						}
					}
				}
			} elseif (GET_EXT_VERSION("sql_patches") == "") {
				// Remove SQLs if extension is not installed
				$SQLs = array();
			}
		} // END - for

		if (!$dry_run) {
			// In normal mode insert task and update extension's version...
			$ext_subj = "[UPDATE-".$ext."-".$EXT_VERSION.":] ".ADMIN_UPDATE_EXT_SUBJ;

			// Create task
			CREATE_EXTENSION_UPDATE_TASK(GET_ADMIN_ID(get_session('admin_login')), $ext_subj, addslashes($NOTES));

			// Update extension's version
			$result = SQL_QUERY_ESC("UPDATE "._MYSQL_PREFIX."_extensions SET ext_version='%s' WHERE ext_name='%s' LIMIT 1",
			 array($EXT_VERSION, $ext), __FILE__, __LINE__);

			// Update cache
			if (EXT_IS_ACTIVE("cache")) {
				if ($cacheInstance->loadCacheFile("extensions", true)) $cacheInstance->destroyCacheFile();
				if ($cacheInstance->loadCacheFile("config")) $cacheInstance->destroyCacheFile();
				if ($cacheInstance->loadCacheFile("mod_reg")) $cacheInstance->destroyCacheFile();
			} // END - if

			// Remove array
			unset($SQLs);
		} else {
			// In "dry-run" mode return array with SQL commands
			return $SQLs;
		}
	}
}

// Output verbose SQL table for extension
function EXTENSION_VERBOSE_TABLE($SQLs, $title = ADMIN_SQLS_EXECUTED_ON_REMOVAL, $dashed = "", $switch = false, $WIDTH = "480") {
	global $_CONFIG;

	$S = false; $SW = 2; $i = 1;
	$OUT = "";
	if ((is_array($SQLs)) && (GET_EXT_VERSION("sql_patches") >= "0.0.7") && ($_CONFIG['verbose_sql'] == "Y")) {
		$OUT  = "<DIV align=\"center\">
<TABLE border=\"0\" cellspacing=\"0\" cellpadding=\"0\" width=\"".$WIDTH."\" align=\"center\"".$dashed.">
<TR>
  <TD colspan=\"2\" align=\"center\" class=\"admin_title bottom2\" height=\"24\">
    <STRONG>".$title.":</STRONG>
  </TD>
</TR>\n";
		foreach ($SQLs as $idx => $sql) {
			$sql = trim($sql);
			if (!empty($sql)) {
				$S = true;
				$OUT .= "<TR>
  <TD class=\"switch_sw".(3 - $SW)." bottom2\" style=\"padding-left: 2px\" width=\"30\">".$i.".</TD>
  <TD class=\"switch_sw".(3 - $SW)." bottom2\" style=\"padding-left: 5px; padding-right: 5px\">
    ".$sql."
  </TD>
</TR>\n";
				if ($switch) $SW = 3 - $SW;
				$i++;
			}
		}
	}

	if ((!$S) && (GET_EXT_VERSION("sql_patches") >= "0.0.7") && ($_CONFIG['verbose_sql'] == "Y")) {
		// No addional SQL commands to run
		$OUT .= "<TR>
  <TD colspan=\"2\" align=\"center\" class=\"switch_sw2 bottom2\" height=\"24\">
    <FONT class=\"admin_note\">".ADMIN_NO_ADDIONAL_SQLS."</FONT>
  </TD>
</TR>\n";
	}

	if (!empty($OUT)) {
		// Add missing close-table tag
		$OUT .= "</TABLE>
</DIV>\n";
	}

	// Return output
	return $OUT;
}

// Get extension name from id
function GET_EXT_NAME ($id) {
	$ret = "";
	global $cacheArray, $_CONFIG;
	if (!empty($cacheArray['extensions']['ext_name'][$id])) {
		// Load from cache
		$ret = $cacheArray['extensions']['ext_name'][$id];

		// Count cache hits
		if (isset($_CONFIG['cache_hits'])) { $_CONFIG['cache_hits']++; } else { $_CONFIG['cache_hits'] = 1; }
	} elseif (!EXT_IS_ACTIVE("cache")) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT ext_name FROM "._MYSQL_PREFIX."_extensions WHERE id=%s LIMIT 1",
		 array(bigintval($id)), __FILE__, __LINE__);
		list($ret) = SQL_FETCHROW($result);
		SQL_FREERESULT($result);
	}
	return $ret;
}

// Get extension id from name
function GET_EXT_ID($name) {
	$ret = 0;
	global $cacheArray, $_CONFIG;
	if (isset($cacheArray['extensions']['ext_id'][$name])) {
		// Load from cache
		$ret = $cacheArray['extensions']['ext_id'][$name];

		// Count cache hits
		if (isset($_CONFIG['cache_hits'])) { $_CONFIG['cache_hits']++; } else { $_CONFIG['cache_hits'] = 1; }
	} elseif (!EXT_IS_ACTIVE("cache")) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT id FROM "._MYSQL_PREFIX."_extensions WHERE ext_name='%s' LIMIT 1",
			array($name), __FILE__, __LINE__);
		list($ret) = SQL_FETCHROW($result);
		SQL_FREERESULT($result);
	}

	// Return value
	return $ret;
}

// Activate given extension
function ACTIVATE_EXTENSION ($ext_name) {
	// Activate the extension
	$result = SQL_QUERY_ESC("UPDATE "._MYSQL_PREFIX."_extensions SET ext_active='Y' WHERE ext_name='%s' LIMIT 1",
		array($ext_name), __FILE__, __LINE__);

	// Extension has been activated?
	if (SQL_AFFECTEDROWS() == 1) {
		// Then run all queries
		EXTENSION_RUN_SQLS(GET_EXT_ID($ext_name), "activate");
	} // END - if
}

// Deactivate given extension
function DEACTIVATE_EXTENSION($ext_name) {
	// Activate the extension
	$result = SQL_QUERY_ESC("UPDATE "._MYSQL_PREFIX."_extensions SET ext_active='N' WHERE ext_name='%s' LIMIT 1",
		array($ext_name), __FILE__, __LINE__);

	// Extension has been activated?
	if (SQL_AFFECTEDROWS() == 1) {
		// Then run all queries
		EXTENSION_RUN_SQLS(GET_EXT_ID($ext_name), "deactivate");

		// Create new task
		CREATE_EXTENSION_DEACTIVATION_TASK($ext_name);

		// Notify the admin
		SEND_ADMIN_NOTIFICATION(ADMIN_SUBJECT_EXTENSION_DEACTIVATED, "admin_ext_deactivated", array('ext_name' => $ext_name));
	} // END - if
}

// Checks wether the extension is older than given
function EXT_VERSION_IS_OLDER ($ext_name, $ext_ver) {
	// Get current extension version
	$currVersion = GET_EXT_VERSION($ext_name);

	// Remove all dots from both versions
	$currVersion = str_replace(".", "", $currVersion);
	$ext_ver = str_replace(".", "", $ext_ver);

	// Now compare both and return the result
	return ($currVersion < $ext_ver);
}

// Creates a new task for updated extension
function CREATE_EXTENSION_UPDATE_TASK ($admin_id, $subject, $notes) {
	// Check if task is not there
	$result = SQL_QUERY_ESC("SELECT id FROM "._MYSQL_PREFIX."_task_system WHERE subject='%s' LIMIT 1",
		array($subject), __FILE__, __LINE__);
	if (SQL_NUMROWS($result) == 0) {
		// Task not created so it's a brand-new extension which we need to register and create a task for!
		$result = SQL_QUERY_ESC("INSERT INTO "._MYSQL_PREFIX."_task_system (assigned_admin, userid, status, task_type, subject, text, task_created) VALUES ('%s','0','NEW','EXTENSION_UPDATE','%s','%s', UNIX_TIMESTAMP())",
			array($admin_id, $subject, $notes), __FILE__, __LINE__);
	} // END - if

	// Free memory
	SQL_FREERESULT($result);
}

// Creates a new task for newly installed extension
function CREATE_NEW_EXTENSION_TASK ($admin_id, $subject, $ext) {
	// Not installed and do we have created a task for the admin?
	$result = SQL_QUERY_ESC("SELECT `id` FROM `"._MYSQL_PREFIX."_task_system` WHERE `subject` LIKE '%s%%' LIMIT 1",
		array($subject), __FILE__, __LINE__);
	if ((SQL_NUMROWS($result) == 0) && (GET_EXT_VERSION($ext) == "")) {
		// Template file
		$tpl = sprintf("%stemplates/%s/html/ext/ext_%s.tpl",
			PATH,
			GET_LANGUAGE(),
			$ext
		);

		// Load text for task
		if (FILE_READABLE($tpl)) {
			// Load extension's own text template (HTML!)
			$msg = LOAD_TEMPLATE("ext_".$ext, true);
		} else {
			// Load default message
			$msg = LOAD_TEMPLATE("admin_new_ext", "", 0);
		}

		// Task not created so it's a brand-new extension which we need to register and create a task for!
		$result_insert = SQL_QUERY_ESC("INSERT INTO "._MYSQL_PREFIX."_task_system (assigned_admin, userid, status, task_type, subject, text, task_created)
VALUES (%s,0,'NEW','EXTENSION','%s','%s',UNIX_TIMESTAMP())",
			array(
				$admin_id,
				$subject,
				addslashes($msg),
			),  __FILE__, __LINE__, true, false
		);
	} // END - if

	// Free memory
	SQL_FREERESULT($result);
}

// Creates a task for automatically deactivated (deprecated) extension
function CREATE_EXTENSION_DEACTIVATION_TASK ($ext) {
	// Create subject line
	$subject = sprintf("[%s:] %s", $ext, TASK_SUBJ_EXTENSION_DEACTIVATED);

	// Not installed and do we have created a task for the admin?
	$result = SQL_QUERY_ESC("SELECT `id` FROM `"._MYSQL_PREFIX."_task_system` WHERE `subject` = '%s' LIMIT 1",
		array($subject), __FILE__, __LINE__);
	if ((SQL_NUMROWS($result) == 0) && (GET_EXT_VERSION($ext) != "")) {
		// Task not created so add it
		$result_insert = SQL_QUERY_ESC("INSERT INTO "._MYSQL_PREFIX."_task_system (assigned_admin, userid, status, task_type, subject, text, task_created)
VALUES (0,0,'NEW','EXTENSION_DEACTIVATION','%s','%s',UNIX_TIMESTAMP())",
			array(
				$subject,
				addslashes(LOAD_TEMPLATE("task_ext_deactivated", true, $ext)),
			),  __FILE__, __LINE__, true, false
		);
	} // END - if

	// Free memory
	SQL_FREERESULT($result);
}

//
?>
